'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var redux = require('redux');

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function strictEqualityCheck(prev, next) {
  return prev === next;
}
function compareArguments(length, prev, next, isEqual) {
  if (!prev) {
    return false;
  }

  for (var i = 0; i < length; i++) {
    if (!isEqual(prev[i], next[i])) {
      return false;
    }
  }

  return true;
}
function memorize(func, equalityCheck) {
  if (equalityCheck === void 0) {
    equalityCheck = strictEqualityCheck;
  }

  var length = func.length;
  var lastArgs;
  var lastResult;
  return function () {
    var nextArgs = arguments;

    if (!compareArguments(length, lastArgs, nextArgs, equalityCheck)) {
      lastResult = func.apply(null, nextArgs);
    }

    lastArgs = nextArgs;
    return lastResult;
  };
}

/**
 * The key used for mounting the orchestrated state subtree on compatible mode
 *
 * @private
 */

var key = '__redux_orchestrate__';
/**
 * Helper - Initialzes store with an empty object.
 * Its main purpose is to notify the registry whether it should enable the compatible mode or not
 *
 * @param {*} state
 */

function defaultReducer(state) {
  if (state === void 0) {
    state = {};
  }

  return state;
}
/**
 * Helper - Resolves restuctured state tree if found
 *
 * @param {*} state
 */

function resolveState(state) {
  if (!state || typeof state[key] === 'undefined') {
    return {
      original: state
    };
  }

  return redux.compose(JSON.parse, JSON.stringify)(state);
}
/**
 * Helper - Restucture state by lifting the original state to top
 *
 * @param {*} state
 */

function restructureState(state) {
  var _Object$assign;

  return Object.assign({}, state.original, (_Object$assign = {}, _Object$assign[key] = state.orchestrated, _Object$assign.toJSON = function toJSON() {
    return state;
  }, _Object$assign));
}
/**
 * Helper - Makes new action creator for the received action type
 *
 * @param {*} type
 */

function makeActionCreator(type) {
  var actionCreator = function actionCreator() {
    for (var _len = arguments.length, payload = new Array(_len), _key = 0; _key < _len; _key++) {
      payload[_key] = arguments[_key];
    }

    return {
      type: type,
      payload: payload
    };
  };

  return Object.assign(actionCreator, {
    /**
     *
     */
    toString: function toString() {
      return type;
    }
  });
}
/**
 * Helper - Makes reducer based on the provided initialize function and the registered event creators
 *
 * @param {*} initialize
 * @param {*} eventCreatorByType
 * @param {*} preloadedState
 */

function makeReducer(initialize, eventCreatorByType, preloadedState) {
  return function (state, action) {
    if (state === void 0) {
      state = preloadedState;
    }

    var eventCreator = eventCreatorByType[action.type];

    if (typeof eventCreator === 'undefined') {
      return initialize(state, action);
    }

    return eventCreator.apply(null, action.payload).call(null, state);
  };
}
/**
 * Helper - Enable migration of store state when redux-orchestrate life up the state tree
 *
 * @param {*} key
 * @param {*} storeState
 * @param {*} reducer
 */

function mirgrateLiftedState(key, storeState, reducer) {
  return function (state, action) {
    if (state === storeState) {
      var _reducer;

      return reducer((_reducer = {}, _reducer[key] = storeState, _reducer), action);
    }

    return reducer(state, action);
  };
}
/**
 * Helper - Liftup event creator from local scope to global scope
 *
 * @param {*} key
 */

function liftUp(key) {
  return function (reducer) {
    return function (state) {
      var _extends2;

      var prevState = state[key];
      var nextState = reducer(prevState);

      if (nextState === prevState) {
        return state;
      }

      return _extends({}, state, (_extends2 = {}, _extends2[key] = nextState, _extends2));
    };
  };
}
/**
 * Helper - Map over all keys by applying the provided function
 *
 * @param {*} obj
 * @param {*} fn
 */

function mapKeys(obj, fn) {
  return Object.keys(obj).reduce(function (result, key) {
    var _Object$assign2;

    return Object.assign(result, (_Object$assign2 = {}, _Object$assign2[fn(key, obj[key])] = obj[key], _Object$assign2));
  }, {});
}
/**
 * Helper - Map over all values by applying the provided function
 *
 * @param {*} obj
 * @param {*} fn
 */

function mapValues(obj, fn) {
  return Object.keys(obj).reduce(function (result, key) {
    var _Object$assign3;

    return Object.assign(result, (_Object$assign3 = {}, _Object$assign3[key] = fn(obj[key], key), _Object$assign3));
  }, {});
}
/**
 * Creates a registry which manages reducers and action creators of the store
 *
 * @param {*} reducer
 * @param {*} emitChange
 */

function createRegistry(store, reducer, preloadedState) {
  var initializeByName = {};
  var eventCreatorByType = {};
  var actionCreatorByService = {};
  var baseReducer = reducer;
  var memorizedRestructureState = memorize(restructureState);
  var isLiftedUp = false;
  /**
   *
   */

  function _replaceReducer() {
    var reducer = baseReducer;

    if (Object.keys(initializeByName).length > 0) {
      reducer = makeReducer(redux.combineReducers(initializeByName), eventCreatorByType, preloadedState);
    }

    if (baseReducer !== defaultReducer) {
      reducer = redux.combineReducers({
        original: baseReducer,
        orchestrated: reducer
      });

      if (!isLiftedUp) {
        isLiftedUp = true;
        reducer = mirgrateLiftedState('original', store.getState(), reducer);
      }
    }

    store.replaceReducer(reducer);
  }

  return {
    initialize: function initialize(name, initialState, eventCreatorByName) {
      var initialize = function initialize(state) {
        if (state === void 0) {
          state = initialState;
        }

        return state;
      };

      var readModel = redux.compose(initialize, function (state) {
        return state[name];
      });
      var eventCreators = mapValues(eventCreatorByName, function (eventCreator) {
        return redux.compose(liftUp(name), eventCreator);
      });
      initializeByName[name] = initialize;
      return {
        readModel: readModel,
        eventCreators: eventCreators
      };
    },
    replaceReducer: function replaceReducer(reducer) {
      if (baseReducer === defaultReducer) {
        return;
      }

      if (reducer === defaultReducer || reducer === baseReducer) {
        return;
      }

      baseReducer = reducer;

      _replaceReducer();
    },
    restructureState: function restructureState(state) {
      if (isLiftedUp) {
        return memorizedRestructureState(state);
      }

      return state;
    },
    register: function register(serviceName, eventCreatorByName) {
      if (typeof actionCreatorByService[serviceName] === 'undefined') {
        eventCreatorByType = Object.assign(eventCreatorByType, mapKeys(eventCreatorByName, function (name) {
          return serviceName + "/" + name;
        }));
        actionCreatorByService[serviceName] = mapValues(eventCreatorByName, function (_, name) {
          return makeActionCreator(serviceName + "/" + name);
        });

        _replaceReducer();
      }

      return actionCreatorByService[serviceName];
    },
    select: function select(state) {
      if (isLiftedUp) {
        return state[key];
      }

      return state;
    }
  };
}
/**
 * Create a storeEnhancer by apply the registry created
 *
 * @param {*} extraArgument
 */

function installRegistry(extraArgument) {
  return function (createStore) {
    return function (reducer, preloadedState) {
      var _resolveState = resolveState(preloadedState),
          original = _resolveState.original,
          orchestrated = _resolveState.orchestrated;

      var store = createStore(reducer, original);
      var registry = createRegistry(store, reducer, orchestrated);
      var getStoreState = store.getState;
      return _extends({}, store, {
        getState: function getState() {
          var storeState = getStoreState();
          var finalState = registry.restructureState(storeState);
          return finalState;
        },
        initialize: function initialize(service) {
          return service(registry, extraArgument);
        },
        replaceReducer: function replaceReducer(reducer) {
          registry.replaceReducer(reducer);
        },
        select: function select(state) {
          return registry.select(state);
        }
      });
    };
  };
}

/**
 * Creates an aggreate of a certain domain
 *
 * @param {*} name
 * @param {*} initialState
 * @param {*} eventCreatorByName
 */

function createAggregate(name, initialState, eventCreatorByName) {
  function aggregate(registry) {
    var _registry$initialize = registry.initialize(name, initialState, eventCreatorByName),
        _readModel = _registry$initialize.readModel,
        eventCreators = _registry$initialize.eventCreators;

    var actionCreators = registry.register("aggregate/" + name, eventCreators);
    return {
      readModel: function readModel(state) {
        return _readModel(state);
      },
      makeCommands: function makeCommands(dispatch) {
        return redux.bindActionCreators(actionCreators, dispatch);
      },
      getEventCreators: function getEventCreators() {
        return _extends({}, eventCreators);
      },
      getSideEffects: function getSideEffects() {
        return {};
      }
    };
  }

  return Object.assign(memorize(aggregate), {
    displayName: name
  });
}

/**
 * Helper - Applying events one by one
 * Returns an identity function if no events
 *
 * @param {*} events
 */

function combineEvents(events) {
  if (events === void 0) {
    events = [];
  }

  return redux.compose.apply(null, [].concat(events).reverse());
}
/**
 * Helper - make the `createEvent` helper using the provided selector
 *
 * @param {*} selector
 */

function makeCreateEvent(selector) {
  return function (handler) {
    if (typeof handler !== 'function') {
      return combineEvents(handler);
    }

    return function (state) {
      var model = selector(state);
      var events = handler(model);
      var event = combineEvents(events);
      return event(state);
    };
  };
}
/**
 * Creates a service of a certain context
 *
 * @param {*} name
 * @param {*} dependencies
 * @param {*} selectorFactory
 * @param {*} eventCreatorFactory
 * @param {*} sideEffectFactory
 * @param {*} onInitialize
 */

function createService(name, dependencies, modelFactory, eventCreatorFactory, sideEffectFactory, onInitialize) {
  function makeSelector(containers) {
    var memorizedModelFactory = memorize(modelFactory);

    var selector = function selector(state) {
      var baseModels = containers.map(function (container) {
        return container.readModel(state);
      });
      var model = memorizedModelFactory.apply(void 0, baseModels);
      return model;
    };

    return memorize(selector);
  }

  function makeEventCreators(containers) {
    var baseEventCreators = containers.map(function (container) {
      return container.getEventCreators();
    });

    if (typeof eventCreatorFactory !== 'function') {
      return baseEventCreators.length === 1 ? baseEventCreators[0] : {};
    }

    return eventCreatorFactory.apply(void 0, baseEventCreators);
  }

  function makeSideEffects(containers) {
    var baseSideEffects = containers.map(function (container) {
      return container.getSideEffects();
    });

    if (typeof sideEffectFactory !== 'function') {
      return baseSideEffects.length === 1 ? baseSideEffects[0] : {};
    }

    return sideEffectFactory.apply(void 0, baseSideEffects);
  }

  function service(registry, extraArgument) {
    var containers = [].concat(dependencies).map(function (dependency) {
      return dependency(registry, extraArgument);
    });
    var selector = makeSelector(containers);
    var eventCreators = makeEventCreators(containers);

    if (typeof eventCreators === 'function') {
      eventCreators = redux.compose(eventCreators, makeCreateEvent)(selector);
    }

    var actionCreators = registry.register("service/" + name, eventCreators);
    var sideEffects = makeSideEffects(containers);

    if (typeof sideEffects === 'function') {
      sideEffects = sideEffects(selector, actionCreators);
    }

    if (typeof onInitialize === 'function') {
      actionCreators = onInitialize(selector, actionCreators, sideEffects, extraArgument) || actionCreators;
    }

    return {
      readModel: function readModel(state) {
        return selector(state);
      },
      makeCommands: function makeCommands(dispatch) {
        return redux.bindActionCreators(actionCreators, dispatch);
      },
      getEventCreators: function getEventCreators() {
        return _extends({}, eventCreators);
      },
      getSideEffects: function getSideEffects() {
        return _extends({}, sideEffects);
      }
    };
  }

  return Object.assign(memorize(service), {
    displayName: name
  });
}

exports.installRegistry = installRegistry;
exports.defaultReducer = defaultReducer;
exports.createAggregate = createAggregate;
exports.createService = createService;
exports.memorize = memorize;
